# -*- coding: utf-8 -*-
#
# Copyright (c) 2024, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the LaTeX Project Public License version 1.3c:
# https://www.latex-project.org/lppl.txt
#


from __future__ import annotations

from .version import __version__, __version_info__
from .err import *
from ._latex_config import latex_config
from ._restricted_pathlib import (
    BaseRestrictedPath,
    StringRestrictedPath,
    SafeStringRestrictedPath,
    SafeWriteStringRestrictedPath,
    ResolvedRestrictedPath,
    SafeResolvedRestrictedPath,
    SafeWriteResolvedRestrictedPath,
)
from ._restricted_subprocess import restricted_run



# -*- coding: utf-8 -*-
#
# Copyright (c) 2015-2024, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the BSD 3-Clause License:
# https://opensource.org/license/BSD-3-clause
#


'''
# `fmtversion`:  Simple version variables for Python packages

Converts version information into a string `__version__` and a named tuple
`__version_info__` suitable for Python packages.  The approach is inspired
by PEP 440 and `sys.version_info`:

  * https://www.python.org/dev/peps/pep-0440

  * https://docs.python.org/3/library/sys.html

Versions of the form "major.minor.micro" are supported, with an optional,
numbered dev/alpha/beta/candidate/final/post status.  The module does not
support more complicated version numbers like "1.0b2.post345.dev456", since
this does not fit into a named tuple of the form used by `sys.version_info`.
The code is written as a single module, so that it may be bundled into
packages, rather than needing to be installed as a separate dependency.

Typical usage in a package's `version.py` with a bundled `fmtversion.py`:

```
from .fmtversion import get_version_plus_info
__version__, __version_info__ = get_version_plus_info(1, 2, 0, 'final', 0)
```

Following `sys.version_info`, `get_version_plus_info()` takes arguments
for a five-component version number:  major, minor, micro, release level, and
serial.  The release level may be any of dev, alpha, beta, candidate, final, or
post, or common variations/abbreviations thereof.  All arguments but
release level must be convertable to integers.

If only `__version__` or `__version_info__` is desired, then the functions
`get_version()` or `get_version_info()` may be used instead.  If a micro
version is not needed (`<major>.<minor>.<micro>`), then set the optional
keyword argument `usemicro=False`.  This will omit a micro version from the
string `__version__`, while the named tuple `__version_info__` will still
have a field `micro` that is set to zero to simplify comparisons.  If each
release level will only have one release, then set `useserial=False`.  This
will omit a serial number from the string `__version__`, while the
named tuple `__version_info__` will still have a field `serial` that is set
to zero.


## License

BSD 3-Clause License

Copyright (c) 2015-2024, Geoffrey M. Poore
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* Neither the name of the copyright holder nor the names of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''




import collections


# Version of this module, which will later be converted into a `__version__`
# and `__version_info__` once the necessary functions are created
__version_tuple__ = (1, 2, 0, 'final', 0)




VersionInfo = collections.namedtuple('VersionInfo', ['major', 'minor', 'micro',
                                                     'releaselevel', 'serial'])


def get_version_info(major, minor, micro, releaselevel, serial,
                     usemicro=True, useserial=True):
    '''
    Create a VersionInfo instance suitable for use as `__version_info__`.

    Perform all type and value checking that is needed for arguments; assume
    that no previous checks have been performed.  This allows all checks to be
    centralized in this single function.
    '''
    if not all(isinstance(x, int) or isinstance(x, str)
            for x in (major, minor, micro, serial)):
        raise TypeError('major, minor, micro, and serial must be integers or strings')
    if not isinstance(releaselevel, str):
        raise TypeError('releaselevel must be a string')
    if not all(isinstance(x, bool) for x in (usemicro, useserial)):
        raise TypeError('usemicro and useserial must be bools')

    try:
        major = int(major)
        minor = int(minor)
        micro = int(micro)
        serial = int(serial)
    except ValueError:
        raise ValueError('major, minor, micro, and serial must be convertable to integers')
    if any(x < 0 for x in (major, minor, micro, serial)):
        raise ValueError('major, minor, micro, and serial must correspond to non-negative integers')
    if not usemicro and micro != 0:
        raise ValueError('usemicro=False, but a micro value "{0}" has been set'.format(micro))
    if not useserial and serial != 0:
        raise ValueError('useserial=False, but a serial value "{0}" has been set'.format(serial))

    releaselevel_dict = {'dev': 'dev',
                         'a': 'a', 'alpha': 'a',
                         'b': 'b', 'beta': 'b',
                         'c': 'c', 'rc': 'c',
                         'candidate': 'c', 'releasecandidate': 'c',
                         'pre': 'c', 'preview': 'c',
                         'final': 'final',
                         'post': 'post', 'r': 'post', 'rev': 'post'}
    try:
        releaselevel = releaselevel_dict[releaselevel.lower()]
    except KeyError:
        raise ValueError('Invalid releaselevel "{0}"'.format(releaselevel))
    if releaselevel == 'final' and serial != 0:
        raise ValueError('final release must not have non-zero serial')

    return VersionInfo(major, minor, micro, releaselevel, serial)


def get_version(*args, **kwargs):
    '''
    Create a version string suitable for use as `__version__`.

    Make sure arguments are appropriate, but leave all actual processing and
    value and type checking to `get_version_info()`.
    '''
    usemicro = kwargs.pop('usemicro', True)
    useserial = kwargs.pop('useserial', True)
    if kwargs:
        raise TypeError('Unexpected keyword(s): {0}'.format(', '.join('{0}'.format(k) for k in kwargs)))

    if len(args) == 1:
        version_info = args[0]
        if not isinstance(version_info, VersionInfo):
            raise TypeError('Positional argument must have 5 components, or be a VersionInfo instance')
    elif len(args) == 5:
        version_info = get_version_info(*args, usemicro=usemicro, useserial=useserial)
    else:
        raise TypeError('Positional argument must have 5 components, or be a VersionInfo instance')

    version = '{0}.{1}'.format(version_info.major, version_info.minor)
    if usemicro:
        version += '.{0}'.format(version_info.micro)
    if version_info.releaselevel != 'final':
        if version_info.releaselevel in ('dev', 'post'):
            version += '.{0}'.format(version_info.releaselevel)
        else:
            version += '{0}'.format(version_info.releaselevel)
        if useserial:
            version += '{0}'.format(version_info.serial)

    return version


def get_version_plus_info(*args, **kwargs):
    '''
    Create a tuple consisting of a version string and a VersionInfo instance.
    '''
    usemicro = kwargs.pop('usemicro', True)
    useserial = kwargs.pop('useserial', True)
    if kwargs:
        raise TypeError('Unexpected keyword(s): {0}'.format(', '.join('{0}'.format(k) for k in kwargs)))

    version_info = get_version_info(*args, usemicro=usemicro, useserial=useserial)
    version = get_version(version_info, usemicro=usemicro, useserial=useserial)
    return (version, version_info)




# Now that the required functions exist, process `__version_tuple__` into
# `__version__` and `__version_info__` for the module
__version__, __version_info__ = get_version_plus_info(*__version_tuple__)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2024, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the LaTeX Project Public License version 1.3c:
# https://www.latex-project.org/lppl.txt
#


from __future__ import annotations




class LatexRestrictedError(Exception):
    pass

class LatexConfigError(LatexRestrictedError):
    pass

class SecurityError(LatexRestrictedError):
    pass

class PathSecurityError(SecurityError):
    pass

class SubprocessError(LatexRestrictedError):
    pass

class ExecutableNotFoundError(SubprocessError):
    pass

class UnapprovedExecutableError(SubprocessError, SecurityError):
    pass

class ExecutablePathSecurityError(SubprocessError, PathSecurityError):
    pass




# -*- coding: utf-8 -*-
#
# Copyright (c) 2024, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the LaTeX Project Public License version 1.3c:
# https://www.latex-project.org/lppl.txt
#


from __future__ import annotations

import os
import shutil
import subprocess
import sys
from .err import LatexConfigError
from ._anypath import AnyPath




class LatexConfig(object):
    '''
    Access config settings from `kpsewhich` (TeX Live) or `initexmf` (MiKTeX),
    plus related environment variables.  Also locate files via `kpsewhich`.

    If the environment variable `TEXSYSTEM` is set to `miktex`, then `PATH` is
    searched with `shutil.which()` for an `initexmf` executable with an
    accompanying `kpsewhich`.  This is only guaranteed to give the correct
    executables (and thus return the correct LaTeX config settings) if there
    is only one MiKTeX installation or if the correct MiKTeX installation is
    the first installation on `PATH`.

    Otherwise, TeX Live is assumed.  Under non-Windows operating systems, the
    `SELFAUTOLOC` environment variable is used to locate the correct
    `kpsewhich`.  Under Windows, `SELFAUTOLOC` may be overwritten by the
    `runscript.exe` wrapper for TeX Live shell escape executables, so it
    cannot be trusted.  (If the shell invokes a TeX Live shell escape
    executable in the wrong TeX installation, due to `PATH` precedence, then
    `SELFAUTOLOC` is overwritten with an incorrect value, and this cannot be
    detected.)  Instead, `PATH` is searched for a `tlmgr` executable with
    accompanying `kpsewhich`.  This is only guaranteed to give the correct
    executables (and thus return the correct LaTeX config settings) if there
    is only one TeX Live installation or if the correct TeX Live installation
    is the first installation on `PATH`.

    File read/write permission settings are available via the following
    properties:

      * `can_read_dotfiles`
      * `can_read_anywhere`
      * `can_write_dotfiles`
      * `can_write_anywhere`

    The `*dotfile` settings describe whether files with names starting with a
    dot `.` are allowed to be read/written.  The `*anywhere` settings describe
    whether files anywhere are allowed to be read/written, or only files
    within the current working directory, TEXMFOUTPUT, TEXMF_OUTPUT_DIRECTORY,
    and their subdirectories.  The values of these properties are determined
    from `openout_any` and `openin_any` settings in `texmf.cnf` for TeX Live,
    and from `[Core]AllowUnsafeInputFiles` and `[Core]AllowUnsafeOutputFiles`
    in `miktex.ini` for MiKTeX.

    Shell escape settings are available via the property
    `can_restricted_shell_escape`.  This is based on `shell_escape` in
    `texmf.cnf` for TeX Live, and on `[Core]ShellCommandMode` in `miktex.ini`
    for MiKTeX.

    Python properties and caching are used extensively so that `kpsewhich` and
    `initexmf` subprocesses only run when their output is actually used and
    has not been obtained previously.
    '''

    def __init__(self):
        pass


    _permitted_subprocess_executables = set(['kpsewhich', 'initexmf', 'miktex-kpsewhich'])
    _permitted_subprocess_executables.update([f'{executable}.exe' for executable in _permitted_subprocess_executables])

    _tex_cwd_anypath: AnyPath = AnyPath.cwd()
    _tex_cwd_str: str = str(_tex_cwd_anypath)

    @property
    def tex_cwd(self) -> str:
        # This is a property so that it is more difficult to modify/so that
        # any modifications are more obvious. `*RestrictedPath` classes depend
        # on `.tex_cwd`, so modifying it bypasses security.
        return self._tex_cwd_str

    _prohibited_subprocess_executable_roots: set[AnyPath] = set()
    _prohibited_subprocess_executable_roots.add(_tex_cwd_anypath)
    # TeX Live allows `TEXMFOUTPUT` to be set in a `texmf.cnf` config file.
    # In `_init_tex_paths()`, if TeX Live is detected and a `TEXMFOUTPUT`
    # environment variable is not defined, then `kpsewhich` is used to
    # retrieve the value of `TEXMFOUTPUT`, and that value is used to update
    # `_prohibited_subprocess_executable_roots`.
    for env_var in [os.getenv(x) for x in ('TEXMFOUTPUT', 'TEXMF_OUTPUT_DIRECTORY')]:
        if env_var:
            env_var_path = AnyPath(env_var)
            _prohibited_subprocess_executable_roots.add(env_var_path)
            _prohibited_subprocess_executable_roots.add(env_var_path.resolve())

    @classmethod
    def _resolve_and_check_executable(cls, executable_name: str, executable_path: AnyPath) -> AnyPath:
        executable_resolved = executable_path.resolve()
        # There is no check for `executable_path.name`, because
        # `executable_path` is always from `shutil.which(<name>)`.
        if executable_resolved.name not in cls._permitted_subprocess_executables:
            raise LatexConfigError(
                f'Executable "{executable_name}" resolved to "{executable_resolved.as_posix()}", '
                f'but "{executable_resolved.name}" is not one of the permitted executables '
                'for determining LaTeX configuration'
            )
        # Executable path can't be writable by LaTeX.  LaTeX writable path(s)
        # can't be relative to the executable's parent directory, since that
        # could make some of the executable's resources writable.  Path
        # comparisons are performed with all permutations of resolved and
        # unresolved paths to reduce the potential for symlink trickery.
        #
        # Fully eliminating the potential for symlink trickery would require
        # checking all locations writable by LaTeX for symlinks to problematic
        # locations.  That isn't worth doing since TeX path security does not
        # consider symlinks.
        #
        # Initially, this doesn't check for the scenario where `TEXMFOUTPUT`
        # is set in a `texmf.cnf` config file with TeX Live.  There isn't a
        # good way to check for that without running `kpsewhich`.  As part of
        # `_init_tex_paths()`, `_prohibited_subprocess_executable_roots` is
        # updated under that scenario by running `kpsewhich` to get
        # `TEXMFOUTPUT`.  Then this function is invoked to check the
        # `kpsewhich` executable against `TEXMFOUTPUT`.  Giving an error
        # message after already running `kpsewhich` isn't ideal, but there
        # isn't a good alternative.  Also, if `TEXMFOUTPUT` is set to an
        # unsafe value in a `texmf.cnf` config file, then all TeX-related
        # executables are potentially compromised, so running `kpsewhich` has
        # a negligible impact on overall security.
        if any(e.is_relative_to(p) or p.is_relative_to(e)
               for e in set([executable_path.parent, executable_resolved.parent])
               for p in cls._prohibited_subprocess_executable_roots):
            raise LatexConfigError(
                f'Executable "{executable_name}" is located under the current directory, TEXMFOUTPUT, or '
                'TEXMF_OUTPUT_DIRECTORY, or one of these locations is under the same directory as the executable'
            )
        return executable_resolved


    _did_init_tex_paths: bool = False
    _texlive_bin: str | None = None
    _texlive_kpsewhich: str | None = None
    _miktex_bin: str | None
    _miktex_initexmf: str | None = None
    _miktex_kpsewhich: str | None = None

    @classmethod
    def _init_tex_paths(cls):
        env_TEXSYSTEM = os.getenv('TEXSYSTEM')
        if env_TEXSYSTEM and env_TEXSYSTEM.lower() == 'miktex':
            cls._init_tex_paths_miktex()
        else:
            cls._init_tex_paths_texlive()

    @classmethod
    def _init_tex_paths_miktex(cls):
        if sys.platform == 'win32':
            which_initexmf = shutil.which('initexmf.exe')
        else:
            which_initexmf = shutil.which('initexmf')
        if not which_initexmf:
            raise LatexConfigError(
                'Environment variable TEXSYSTEM="miktex", but failed to find "initexmf" executable on PATH'
            )
        which_initexmf_path = AnyPath(which_initexmf)
        which_initexmf_resolved = cls._resolve_and_check_executable('initexmf', which_initexmf_path)
        miktex_bin_path = which_initexmf_resolved.parent
        if sys.platform == 'win32':
            which_kpsewhich = shutil.which('kpsewhich.exe', path=str(miktex_bin_path))
        else:
            which_kpsewhich = shutil.which('kpsewhich', path=str(miktex_bin_path))
            if not which_kpsewhich:
                which_kpsewhich = shutil.which('miktex-kpsewhich', path=str(miktex_bin_path))
        if not which_kpsewhich:
            raise LatexConfigError(
                'Environment variable TEXSYSTEM="miktex", '
                'but failed to find an "initexmf" executable with accompanying "kpsewhich" on PATH'
            )
        which_kpsewhich_path = AnyPath(which_kpsewhich)
        which_kpsewhich_resolved = cls._resolve_and_check_executable('kpsewhich', which_kpsewhich_path)
        if not miktex_bin_path == which_kpsewhich_resolved.parent:
            raise LatexConfigError(
                'Environment variable TEXSYSTEM="miktex", '
                f'but "initexmf" executable from PATH resolved to "{which_initexmf_resolved.as_posix()}" '
                f'while "kpsewhich" resolved to "{which_kpsewhich_resolved.as_posix()}"; '
                '"initexmf" and "kpsewhich" should be in the same location'
            )
        cls._miktex_bin = str(miktex_bin_path)
        cls._miktex_initexmf = str(which_initexmf_resolved)
        cls._miktex_kpsewhich = str(which_kpsewhich_resolved)
        cls._did_init_tex_paths = True

    @classmethod
    def _init_tex_paths_texlive(cls):
        env_SELFAUTOLOC = os.getenv('SELFAUTOLOC')
        if not env_SELFAUTOLOC:
            raise LatexConfigError('Environment variable SELFAUTOLOC is expected for TeX Live, but was not set')
        if sys.platform == 'win32':
            # Under Windows, shell escape executables installed within TeX
            # Live are launched with the `runscript.exe` wrapper.  This uses
            # `kpathsea` internally, which will overwrite any existing value
            # of `SELFAUTOLOC`.  As a result, under Windows, `SELFAUTOLOC` may
            # only give the location of the wrapper executable rather than the
            # location of the TeX executable that is invoking shell escape.
            which_tlmgr = shutil.which('tlmgr')  # No `.exe`; likely `.bat`
            if not which_tlmgr:
                raise LatexConfigError('Failed to find TeX Live "tlmgr" executable on PATH')
            which_tlmgr_resolved = AnyPath(which_tlmgr).resolve()
            texlive_bin_path = which_tlmgr_resolved.parent
            which_kpsewhich = shutil.which('kpsewhich.exe', path=str(texlive_bin_path))
            if not which_kpsewhich:
                raise LatexConfigError(
                    'Failed to find a TeX Live "tlmgr" executable with accompanying "kpsewhich" executable on PATH'
                )
            which_kpsewhich_path = AnyPath(which_kpsewhich)
            which_kpsewhich_resolved = cls._resolve_and_check_executable('kpsewhich', which_kpsewhich_path)
            if not texlive_bin_path == which_kpsewhich_resolved.parent:
                raise LatexConfigError(
                    f'"tlmgr" executable from PATH resolved to "{which_tlmgr_resolved.as_posix()}" '
                    f'while "kpsewhich" resolved to "{which_kpsewhich_resolved.as_posix()}"; '
                    '"tlmgr" and "kpsewhich" should be in the same location'
                )
        else:
            env_SELFAUTOLOC_which_kpsewhich = shutil.which('kpsewhich', path=env_SELFAUTOLOC)
            if not env_SELFAUTOLOC_which_kpsewhich:
                raise LatexConfigError(
                    f'Environment variable SELFAUTOLOC has value "{env_SELFAUTOLOC}", '
                    'but a "kpsewhich" executable was not found at that location'
                )
            which_kpsewhich_path = AnyPath(env_SELFAUTOLOC_which_kpsewhich)
            which_kpsewhich_resolved = cls._resolve_and_check_executable('kpsewhich', which_kpsewhich_path)
            texlive_bin_path = which_kpsewhich_resolved.parent
        cls._texlive_bin = str(texlive_bin_path)
        cls._texlive_kpsewhich = str(which_kpsewhich_resolved)
        cls._did_init_tex_paths = True
        if not os.getenv('TEXMFOUTPUT'):
            # With TeX Live, check for unsafe value of `TEXMFOUTPUT` from
            # `texmf.cnf` config file.
            config_TEXMFOUTPUT = cls._get_texlive_var_value('TEXMFOUTPUT')
            if config_TEXMFOUTPUT:
                config_TEXMFOUTPUT_path = AnyPath(config_TEXMFOUTPUT)
                cls._prohibited_subprocess_executable_roots.add(config_TEXMFOUTPUT_path)
                cls._prohibited_subprocess_executable_roots.add(config_TEXMFOUTPUT_path.resolve())
                cls._resolve_and_check_executable('kpsewhich', which_kpsewhich_path)


    # Locations of TeX executables must be returned as strings, not as
    # `AnyPath`.  All non-private paths should be subclasses of
    # `RestrictedPath`, but it can't be defined without `LatexConfig`.

    @property
    def texlive_bin(self) -> str | None:
        if not self._did_init_tex_paths:
            self._init_tex_paths()
        return self._texlive_bin

    @property
    def texlive_kpsewhich(self) -> str | None:
        if not self._did_init_tex_paths:
            self._init_tex_paths()
        return self._texlive_kpsewhich

    @property
    def miktex_bin(self) -> str | None:
        if not self._did_init_tex_paths:
            self._init_tex_paths()
        return self._miktex_bin

    @property
    def miktex_initexmf(self) -> str | None:
        if not self._did_init_tex_paths:
            self._init_tex_paths()
        return self._miktex_initexmf

    @property
    def miktex_kpsewhich(self) -> str | None:
        if not self._did_init_tex_paths:
            self._init_tex_paths()
        return self._miktex_kpsewhich


    def kpsewhich_find_config_file(self, file: str,) -> str | None:
        if not self._did_init_tex_paths:
            self._init_tex_paths()
        if self.texlive_kpsewhich:
            kpsewhich = self.texlive_kpsewhich
        elif self.miktex_kpsewhich:
            kpsewhich = self.miktex_kpsewhich
        else:
            raise TypeError
        cmd = [kpsewhich, '-f', 'othertext', file]
        proc = subprocess.run(cmd, shell=False, capture_output=True)
        value = proc.stdout.strip().decode(sys.stdout.encoding) or None
        return value

    _kpsewhich_find_file_cache: dict[str, str | None] = {}

    def kpsewhich_find_file(self, file: str, *, cache: bool = False) -> str | None:
        if cache:
            try:
                return self._kpsewhich_find_file_cache[file]
            except KeyError:
                pass
        if not self._did_init_tex_paths:
            self._init_tex_paths()
        if self.texlive_kpsewhich:
            kpsewhich = self.texlive_kpsewhich
        elif self.miktex_kpsewhich:
            kpsewhich = self.miktex_kpsewhich
        else:
            raise TypeError
        cmd = [kpsewhich, file]
        proc = subprocess.run(cmd, shell=False, capture_output=True)
        value = proc.stdout.strip().decode(sys.stdout.encoding) or None
        if cache:
            self._kpsewhich_find_file_cache[file] = value
            return self._kpsewhich_find_file_cache[file]
        return value


    _texlive_var_value_cache: dict[str, str | None] = {}

    @classmethod
    def _get_texlive_var_value(cls, var: str) -> str | None:
        try:
            return cls._texlive_var_value_cache[var]
        except KeyError:
            pass
        if cls._texlive_kpsewhich is None:
            raise TypeError
        cmd = [cls._texlive_kpsewhich, '--var-value', var]
        proc = subprocess.run(cmd, shell=False, capture_output=True)
        value = proc.stdout.strip().decode(sys.stdout.encoding) or None
        if var.lower() in ('openin_any', 'openout_any'):
            # Documentation for `openin_any` and `openout_any` values:
            # https://www.tug.org/texinfohtml/kpathsea.html#Safe-filenames-1
            if value:
                value = value.lower()
            if value in ('y', '1'):
                value = 'a'
            elif value in ('n', '0'):
                value = 'r'
            elif value not in ('a', 'r', 'p'):
                value = 'p'
        elif var.lower() == 'shell_escape_commands':
            if value:
                value = value.rstrip(',%')
        cls._texlive_var_value_cache[var] = value
        return cls._texlive_var_value_cache[var]

    _miktex_config_value_cache: dict[str, str | None] = {}

    @classmethod
    def _get_miktex_config_value(cls, var: str) -> str | None:
        try:
            return cls._miktex_config_value_cache[var]
        except KeyError:
            pass
        if cls._miktex_initexmf is None:
            raise TypeError
        cmd = [cls._miktex_initexmf, '--show-config-value', var]
        proc = subprocess.run(cmd, shell=False, capture_output=True)
        value = proc.stdout.strip().decode(sys.stdout.encoding) or None
        if var.lower() in ('[core]allowunsafeinputfiles', '[core]allowunsafeoutputfiles'):
            if value:
                value = value.lower()
            if value not in ('true', 'false'):
                value = 'false'
        cls._miktex_config_value_cache[var] = value
        return cls._miktex_config_value_cache[var]


    _did_init_read_settings: bool = False
    _can_read_dotfiles: bool = False
    _can_read_anywhere: bool = False

    _did_init_write_settings: bool = False
    _can_write_dotfiles: bool = False
    _can_write_anywhere: bool = False

    # Documentation for `openin_any` and `openout_any` values:
    # https://www.tug.org/texinfohtml/kpathsea.html#Safe-filenames-1

    @classmethod
    def _init_read_settings(cls):
        if not cls._did_init_tex_paths:
            cls._init_tex_paths()
        if cls._texlive_kpsewhich:
            openin_any = cls._get_texlive_var_value('openin_any')
            if openin_any == 'a':
                cls._can_read_dotfiles = True
                cls._can_read_anywhere = True
            elif openin_any == 'r':
                cls._can_read_dotfiles = False
                cls._can_read_anywhere = True
            elif openin_any == 'p':
                cls._can_read_dotfiles = False
                cls._can_read_anywhere = False
            else:
                raise ValueError
        elif cls._miktex_initexmf:
            allow_unsafe_input_files = cls._get_miktex_config_value('[Core]AllowUnsafeInputFiles')
            if allow_unsafe_input_files == 'true':
                cls._can_read_dotfiles = True
                cls._can_read_anywhere = True
            elif allow_unsafe_input_files == 'false':
                cls._can_read_dotfiles = False
                cls._can_read_anywhere = False
            else:
                raise ValueError
        else:
            raise TypeError
        cls._did_init_read_settings = True

    @property
    def can_read_dotfiles(self) -> bool:
        if not self._did_init_read_settings:
            self._init_read_settings()
        return self._can_read_dotfiles

    @property
    def can_read_anywhere(self) -> bool:
        if not self._did_init_read_settings:
            self._init_read_settings()
        return self._can_read_anywhere

    @classmethod
    def _init_write_settings(cls):
        if not cls._did_init_tex_paths:
            cls._init_tex_paths()
        if cls._texlive_kpsewhich:
            openout_any = cls._get_texlive_var_value('openout_any')
            if openout_any == 'a':
                cls._can_write_dotfiles = True
                cls._can_write_anywhere = True
            elif openout_any == 'r':
                cls._can_write_dotfiles = False
                cls._can_write_anywhere = True
            elif openout_any == 'p':
                cls._can_write_dotfiles = False
                cls._can_write_anywhere = False
            else:
                raise ValueError
        elif cls._miktex_initexmf:
            allow_unsafe_output_files = cls._get_miktex_config_value('[Core]AllowUnsafeOutputFiles')
            if allow_unsafe_output_files == 'true':
                cls._can_write_dotfiles = True
                cls._can_write_anywhere = True
            elif allow_unsafe_output_files == 'false':
                cls._can_write_dotfiles = False
                cls._can_write_anywhere = False
            else:
                raise TypeError
        else:
            raise TypeError
        cls._did_init_write_settings = True

    @property
    def can_write_dotfiles(self) -> bool:
        if not self._did_init_write_settings:
            self._init_write_settings()
        return self._can_write_dotfiles

    @property
    def can_write_anywhere(self) -> bool:
        if not self._did_init_write_settings:
            self._init_write_settings()
        return self._can_write_anywhere

    _prohibited_write_file_extensions: frozenset[str] | None
    # Microsoft default PATHEXT and kpathsea default fallback PATHEXT are
    # slightly different, so use a union of the two.
    # https://learn.microsoft.com/en-us/windows-server/administration/windows-commands/start
    _microsoft_default_pathext = '.COM;.EXE;.BAT;.CMD;.VBS;.VBE;.JS;.JSE;.WSF;.WSH;.MSC'
    # https://tug.org/svn/texlive/trunk/Build/source/texk/kpathsea/progname.c?revision=57915&view=markup#l415
    _kpathsea_default_pathext = '.com;.exe;.bat;.cmd;.vbs;.vbe;.js;.jse;.wsf;.wsh;.ws;.tcl;.py;.pyw'
    _fallback_prohibited_write_file_extensions = frozenset(
        ';'.join([_microsoft_default_pathext, _kpathsea_default_pathext]).lower().split(';')
    )
    if sys.platform == 'win32':
        _pathext = os.getenv('PATHEXT')
        if _pathext:
            _prohibited_write_file_extensions = frozenset(x for x in _pathext.lower().split(os.pathsep) if x)
        else:
            _prohibited_write_file_extensions = _fallback_prohibited_write_file_extensions
    elif sys.platform == 'cygwin':
        # This follows kpathsea:
        # https://tug.org/svn/texlive/trunk/Build/source/texk/kpathsea/progname.c?revision=57915&view=markup#l424
        _prohibited_write_file_extensions = _fallback_prohibited_write_file_extensions
    else:
        _prohibited_write_file_extensions = None

    @property
    def prohibited_write_file_extensions(self) -> frozenset[str] | None:
        return self._prohibited_write_file_extensions

    _did_init_shell_escape_settings: bool = False
    _can_restricted_shell_escape: bool

    @classmethod
    def _init_shell_escape_settings(cls):
        if not cls._did_init_tex_paths:
            cls._init_tex_paths()
        if cls._texlive_kpsewhich:
            # https://tug.org/svn/texlive/trunk/Build/source/texk/kpathsea/texmf.cnf?revision=70942&view=markup#l634
            shell_escape = cls._get_texlive_var_value('shell_escape')
            if shell_escape and shell_escape.lower() in ('t', 'p'):
                cls._can_restricted_shell_escape = True
            else:
                cls._can_restricted_shell_escape = False
        elif cls._miktex_initexmf:
            # https://docs.miktex.org/manual/miktex.ini.html
            shell_command_mode = cls._get_miktex_config_value('[Core]ShellCommandMode')
            if shell_command_mode and shell_command_mode.lower() in ('restricted', 'unrestricted'):
                cls._can_restricted_shell_escape = True
            else:
                cls._can_restricted_shell_escape = False
        else:
            raise TypeError
        cls._did_init_shell_escape_settings = True

    @property
    def can_restricted_shell_escape(self):
        if not self._did_init_shell_escape_settings:
            self._init_shell_escape_settings()
        return self._can_restricted_shell_escape


    _var_str_none_cache: dict[str, str | None] = {}

    @property
    def TEXMFHOME(self) -> str | None:
        try:
            return self._var_str_none_cache['TEXMFHOME']
        except KeyError:
            if self.texlive_kpsewhich:
                value = self._get_texlive_var_value('TEXMFHOME')
            elif self.miktex_initexmf:
                value = self._get_miktex_config_value('TEXMFHOME')
            else:
                raise TypeError
            self._var_str_none_cache['TEXMFHOME'] = value
            return self._var_str_none_cache['TEXMFHOME']

    @property
    def TEXMFOUTPUT(self) -> str | None:
        try:
            return self._var_str_none_cache['TEXMFOUTPUT']
        except KeyError:
            value = os.getenv('TEXMFOUTPUT')
            if value is None and self.texlive_kpsewhich:
                # TeX Live allows `TEXMFOUTPUT` to be set in `texmf.cnf`
                value = self._get_texlive_var_value('TEXMFOUTPUT')
            self._var_str_none_cache['TEXMFOUTPUT'] = value
            return self._var_str_none_cache['TEXMFOUTPUT']

    @property
    def TEXMF_OUTPUT_DIRECTORY(self) -> str | None:
        try:
            return self._var_str_none_cache['TEXMF_OUTPUT_DIRECTORY']
        except KeyError:
            value = os.getenv('TEXMF_OUTPUT_DIRECTORY')
            self._var_str_none_cache['TEXMF_OUTPUT_DIRECTORY'] = value
            return self._var_str_none_cache['TEXMF_OUTPUT_DIRECTORY']

    _var_frozenset_cache: dict[str, frozenset[str]] = {}

    @property
    def restricted_shell_escape_commands(self) -> frozenset[str]:
        try:
            return self._var_frozenset_cache['restricted_shell_escape_commands']
        except KeyError:
            commands = set()
            if self.texlive_kpsewhich:
                value = self._get_texlive_var_value('shell_escape_commands')
                if value is not None:
                    commands.update(value.split(','))
            elif self.miktex_initexmf:
                value = self._get_miktex_config_value('[Core]AllowedShellCommands[]')
                if value is not None:
                    commands.update(value.split(';'))
            else:
                raise TypeError
            self._var_frozenset_cache['restricted_shell_escape_commands'] = frozenset(commands)
            return self._var_frozenset_cache['restricted_shell_escape_commands']




latex_config = LatexConfig()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2024, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the LaTeX Project Public License version 1.3c:
# https://www.latex-project.org/lppl.txt
#


from __future__ import annotations

import os
from contextlib import AbstractContextManager
from typing import Literal
try:
    from typing import Self
except ImportError:
    pass
from .err import PathSecurityError
from ._anypath import AnyPath
from ._latex_config import latex_config




# Following the approach of `contextlib.chdir()` from Python 3.11+:
# https://github.com/python/cpython/blob/3.12/Lib/contextlib.py#L788-L800
class _CdTeXCwd(AbstractContextManager):
    def __init__(self):
        self._tex_cwd: str = latex_config.tex_cwd
        self._cwd_stack: list[str | None] = []

    def __enter__(self):
        cwd = os.getcwd()
        if cwd == self._tex_cwd:
            self._cwd_stack.append(None)
        else:
            self._cwd_stack.append(cwd)
            os.chdir(self._tex_cwd)

    def __exit__(self, *excinfo):
        last_cwd = self._cwd_stack.pop()
        if last_cwd is not None:
            os.chdir(last_cwd)

_cd_tex_cwd = _CdTeXCwd()




class BaseRestrictedPath(type(AnyPath())):
    '''
    Base class for `RestrictedPath` classes.  This is based on `AnyPath`,
    instead of directly on `pathlib.Path`, for better Python 3.8 compatibility
    and also for `.resolve()` caching.

    Redefines all methods that modify the file system.

      * Most methods for opening, reading, writing, replacing, and deleting
        files as well as methods for creating and deleting directories are
        redefined to depend on the new methods `.readable_dir()`,
        `.readable_file()`, `.writable_dir()`, and `.writable_file()`.  These
        new methods all raise `NotImplementedError` in the base class; they
        must be implemented by subclasses.

      * All other methods for modifying the file system are redefined to raise
        `NotImplementedError`.  Most of these methods relate to file
        permissions and links.

    All supported methods that modify the file system operate with the TeX
    working directory as the current working directory.  This is necessary for
    performing path security analysis.  It is best not to modify the current
    working directory.  However, it is possible to create paths when the
    current working directory is another location and use those paths to
    obtain information from the file system (for example, via `.exists()`).
    Such paths will not work correctly to *modify* the file system; they must
    first be converted into absolute paths or paths relative to the TeX
    working directory.
    '''

    __slots__ = ()


    # Ensure paths are made relative to TeX working directory

    def absolute(self, *args, **kwargs):
        with _cd_tex_cwd:
            return super().absolute(*args, **kwargs)

    def resolve(self, *args, **kwargs):
        with _cd_tex_cwd:
            return super().resolve(*args, **kwargs)


    # This must be defined by subclasses.  If false, the file system is
    # accessed with a path as the path is currently defined.  If true, the
    # path is resolved first.  This should be true for subclasses that perform
    # path security analysis using resolved paths.  Otherwise, the file system
    # (symlinks) could be modified after security analysis but before a path
    # is used, so that the security analysis is invalid.  Note that due to the
    # `.resolve()` caching from `AnyPath`, a path will always resolve in the
    # same way.
    @classmethod
    def access_file_system_with_resolved_paths(cls) -> bool:
        raise NotImplementedError

    # Default security is based on TeX configuration.  Subclasses can override
    # this, but should typically only do so to increase security to maximum.
    @classmethod
    def can_read_anywhere(cls) -> bool:
        return latex_config.can_read_anywhere

    @classmethod
    def can_read_dotfiles(cls) -> bool:
        return latex_config.can_read_dotfiles

    @classmethod
    def can_write_anywhere(cls) -> bool:
        return latex_config.can_write_anywhere

    @classmethod
    def can_write_dotfiles(cls) -> bool:
        return latex_config.can_write_dotfiles

    @classmethod
    def prohibited_write_file_extensions(cls) -> frozenset[str] | None:
        return latex_config.prohibited_write_file_extensions

    # Caches use `self.cache_key` which includes the class, so that the
    # returned type is correct.  Values describe whether paths are accessible,
    # and if not, why:
    #     (is_accessible: bool, reason_not_accessible: str | None)
    _readable_dir_cache:  dict[tuple[type[Self], Self], tuple[Literal[True], None] | tuple[Literal[False], str]] = {}
    _readable_file_cache: dict[tuple[type[Self], Self], tuple[Literal[True], None] | tuple[Literal[False], str]] = {}
    _writable_dir_cache:  dict[tuple[type[Self], Self], tuple[Literal[True], None] | tuple[Literal[False], str]] = {}
    _writable_file_cache: dict[tuple[type[Self], Self], tuple[Literal[True], None] | tuple[Literal[False], str]] = {}

    def readable_dir(self) -> tuple[Literal[True], None] | tuple[Literal[False], str]:
        raise NotImplementedError

    def readable_file(self) -> tuple[Literal[True], None] | tuple[Literal[False], str]:
        raise NotImplementedError

    def writable_dir(self) -> tuple[Literal[True], None] | tuple[Literal[False], str]:
        raise NotImplementedError

    def writable_file(self) -> tuple[Literal[True], None] | tuple[Literal[False], str]:
        raise NotImplementedError


    def chmod(self, *args, **kwargs):
        raise NotImplementedError

    def copy(self, *args, **kwargs):
        # Python 3.14
        raise NotImplementedError

    def copytree(self, *args, **kwargs):
        # Python 3.14
        raise NotImplementedError

    def lchmod(self, *args, **kwargs):
        raise NotImplementedError

    def mkdir(self, *args, **kwargs):
        with _cd_tex_cwd:
            is_writable, reason = self.writable_dir()
            if not is_writable:
                raise PathSecurityError(f'Cannot create directory "{self.as_posix()}":  {reason}')
            if self.access_file_system_with_resolved_paths() and not self.is_resolved():
                return super(BaseRestrictedPath, self.resolve()).mkdir(*args, **kwargs)
            return super().mkdir(*args, **kwargs)

    def open(self, mode: str = 'r', *args, **kwargs):
        with _cd_tex_cwd:
            if 'r' in mode:
                is_readable, reason = self.readable_file()
                if not is_readable:
                    raise PathSecurityError(f'Cannot read file "{self.as_posix()}":  {reason}')
            elif any(char in mode for char in 'wxa'):
                is_writable, reason = self.writable_file()
                if not is_writable:
                    raise PathSecurityError(f'Cannot write file "{self.as_posix()}":  {reason}')
            else:
                raise NotImplementedError
            if self.access_file_system_with_resolved_paths() and not self.is_resolved():
                return super(BaseRestrictedPath, self.resolve()).open(mode=mode, *args, **kwargs)
            return super().open(mode=mode, *args, **kwargs)

    def read_bytes(self, *args, **kwargs):
        with _cd_tex_cwd:
            is_readable, reason = self.readable_file()
            if not is_readable:
                raise PathSecurityError(f'Cannot read file "{self.as_posix()}":  {reason}')
            if self.access_file_system_with_resolved_paths() and not self.is_resolved():
                return super(BaseRestrictedPath, self.resolve()).read_bytes(*args, **kwargs)
            return super().read_bytes(*args, **kwargs)

    def read_text(self, *args, **kwargs) -> str:
        with _cd_tex_cwd:
            is_readable, reason = self.readable_file()
            if not is_readable:
                raise PathSecurityError(f'Cannot read file "{self.as_posix()}":  {reason}')
            if self.access_file_system_with_resolved_paths() and not self.is_resolved():
                return super(BaseRestrictedPath, self.resolve()).read_text(*args, **kwargs)
            return super().read_text(*args, **kwargs)

    def rename(self, target: Self):
        with _cd_tex_cwd:
            is_writable, reason = self.writable_file()
            if not is_writable:
                raise PathSecurityError(f'Cannot rename file "{self.as_posix()}":  {reason}')
            target_is_writable, target_reason = target.writable_file()
            if not target_is_writable:
                raise PathSecurityError(f'Cannot create renamed file "{target.as_posix()}":  {target_reason}')
            if self.access_file_system_with_resolved_paths() and not (self.is_resolved() and target.is_resolved()):
                return super(BaseRestrictedPath, self.resolve()).rename(target.resolve())
            return super().rename(target)

    def replace(self, target: Self):
        with _cd_tex_cwd:
            is_writable, reason = self.writable_file()
            if not is_writable:
                raise PathSecurityError(f'Cannot replace file "{self.as_posix()}":  {reason}')
            target_is_writable, target_reason = target.writable_file()
            if not target_is_writable:
                raise PathSecurityError(f'Cannot create replacement file "{target.as_posix()}":  {target_reason}')
            if self.access_file_system_with_resolved_paths() and not (self.is_resolved() and target.is_resolved()):
                return super(BaseRestrictedPath, self.resolve()).replace(target.resolve())
            return super().replace(target)

    def rmdir(self):
        with _cd_tex_cwd:
            is_writable, reason = self.writable_dir()
            if not is_writable:
                raise PathSecurityError(f'Cannot delete directory "{self.as_posix()}":  {reason}')
            if self.access_file_system_with_resolved_paths() and not self.is_resolved():
                return super(BaseRestrictedPath, self.resolve()).rmdir()
            return super().rmdir()

    def symlink_to(self, *args, **kwargs):
        raise NotImplementedError

    def hardlink_to(self, *args, **kwargs):
        raise NotImplementedError

    def touch(self, *args, **kwargs):
        with _cd_tex_cwd:
            is_writable, reason = self.writable_file()
            if not is_writable:
                raise PathSecurityError(f'Cannot create file "{self.as_posix()}":  {reason}')
            if self.access_file_system_with_resolved_paths() and not self.is_resolved():
                return super(BaseRestrictedPath, self.resolve()).touch(*args, **kwargs)
            return super().touch(*args, **kwargs)

    def unlink(self, *args, **kwargs):
        with _cd_tex_cwd:
            is_writable, reason = self.writable_file()
            if not is_writable:
                raise PathSecurityError(f'Cannot delete file "{self.as_posix()}":  {reason}')
            if self.access_file_system_with_resolved_paths() and not self.is_resolved():
                return super(BaseRestrictedPath, self.resolve()).unlink(*args, **kwargs)
            return super().unlink(*args, **kwargs)

    def write_bytes(self, *args, **kwargs):
        with _cd_tex_cwd:
            is_writable, reason = self.writable_file()
            if not is_writable:
                raise PathSecurityError(f'Cannot write file "{self.as_posix()}":  {reason}')
            if self.access_file_system_with_resolved_paths() and not self.is_resolved():
                return super(BaseRestrictedPath, self.resolve()).write_bytes(*args, **kwargs)
            return super().write_bytes(*args, **kwargs)

    def write_text(self, *args, **kwargs):
        with _cd_tex_cwd:
            is_writable, reason = self.writable_file()
            if not is_writable:
                raise PathSecurityError(f'Cannot write file "{self.as_posix()}":  {reason}')
            if self.access_file_system_with_resolved_paths() and not self.is_resolved():
                return super(BaseRestrictedPath, self.resolve()).write_text(*args, **kwargs)
            return super().write_text(*args, **kwargs)


    @classmethod
    def tex_cwd(cls) -> Self:
        try:
            return cls._tex_cwd
        except AttributeError:
            cls._tex_cwd = cls(latex_config.tex_cwd)
            return cls._tex_cwd

    @classmethod
    def TEXMFOUTPUT(cls) -> Self | None:
        try:
            return cls._TEXMFOUTPUT
        except AttributeError:
            if latex_config.TEXMFOUTPUT is None:
                cls._TEXMFOUTPUT = None
            else:
                cls._TEXMFOUTPUT = cls(latex_config.TEXMFOUTPUT)
            return cls._TEXMFOUTPUT

    @classmethod
    def TEXMF_OUTPUT_DIRECTORY(cls) -> Self | None:
        try:
            return cls._TEXMF_OUTPUT_DIRECTORY
        except AttributeError:
            if latex_config.TEXMF_OUTPUT_DIRECTORY is None:
                cls._TEXMF_OUTPUT_DIRECTORY = None
            else:
                cls._TEXMF_OUTPUT_DIRECTORY = cls(latex_config.TEXMF_OUTPUT_DIRECTORY)
            return cls._TEXMF_OUTPUT_DIRECTORY

    @classmethod
    def tex_openout_roots(cls) -> tuple[Self]:
        # This is a tuple of paths because these are the directories that TeX
        # will try to write to in order.  If TEXMF_OUTPUT_DIRECTORY or the
        # current working directory is not writable, then TeX will fall back
        # to TEXMFOUTPUT if it is set.
        try:
            return cls._tex_openout_roots
        except AttributeError:
            openout_roots = []
            TEXMF_OUTPUT_DIRECTORY = cls.TEXMF_OUTPUT_DIRECTORY()
            if TEXMF_OUTPUT_DIRECTORY:
                openout_roots.append(TEXMF_OUTPUT_DIRECTORY)
            else:
                openout_roots.append(cls.tex_cwd())
            TEXMFOUTPUT = cls.TEXMFOUTPUT()
            if TEXMFOUTPUT and TEXMFOUTPUT not in openout_roots:
                openout_roots.append(TEXMFOUTPUT)
            cls._tex_openout_roots = tuple(openout_roots)
            return cls._tex_openout_roots

    @classmethod
    def tex_roots(cls) -> frozenset[Self]:
        # Root paths of all locations that TeX can always access, even when
        # file system access is restricted ("paranoid" mode).
        try:
            return cls._tex_roots
        except AttributeError:
            tex_roots = set()
            tex_roots.add(cls.tex_cwd())
            TEXMF_OUTPUT_DIRECTORY = cls.TEXMF_OUTPUT_DIRECTORY()
            if TEXMF_OUTPUT_DIRECTORY:
                tex_roots.add(TEXMF_OUTPUT_DIRECTORY)
            TEXMFOUTPUT = cls.TEXMFOUTPUT()
            if TEXMFOUTPUT:
                tex_roots.add(TEXMFOUTPUT)
            cls._tex_roots = frozenset(tex_roots)
            return cls._tex_roots

    @classmethod
    def tex_roots_resolved(cls) -> frozenset[Self]:
        try:
            return cls._tex_roots_resolved
        except AttributeError:
            cls._tex_roots_resolved = frozenset(p.resolve() for p in cls.tex_roots())
            return cls._tex_roots_resolved

    @classmethod
    def tex_roots_with_resolved(cls) -> frozenset[Self]:
        try:
            return cls._tex_roots_with_resolved
        except KeyError:
            cls._tex_roots_with_resolved = cls.tex_roots() | cls.tex_roots_resolved()
            return cls._tex_roots_with_resolved

    @classmethod
    def tex_texmfoutput_roots(cls) -> frozenset[Self]:
        try:
            return cls._tex_texmfoutput_roots
        except AttributeError:
            tex_texmfoutput_roots = set()
            TEXMF_OUTPUT_DIRECTORY = cls.TEXMF_OUTPUT_DIRECTORY()
            if TEXMF_OUTPUT_DIRECTORY:
                tex_texmfoutput_roots.add(TEXMF_OUTPUT_DIRECTORY)
            TEXMFOUTPUT = cls.TEXMFOUTPUT()
            if TEXMFOUTPUT:
                tex_texmfoutput_roots.add(TEXMFOUTPUT)
            cls._tex_texmfoutput_roots = frozenset(tex_texmfoutput_roots)
            return cls._tex_texmfoutput_roots




class StringRestrictedPath(BaseRestrictedPath):
    '''
    Restrict paths by analyzing them as strings.  This follows the approach
    taken in TeX's file system security.  For example, see
    https://www.tug.org/texinfohtml/kpathsea.html#Safe-filenames-1 and
    https://tug.org/svn/texlive/trunk/Build/source/texk/kpathsea/progname.c?revision=57915&view=markup#l414.
    Restrictions depend on LaTeX configuration.  Restrictions are determined
    by `openout_any` and `openin_any` settings in `texmf.cnf` for TeX Live,
    and from `[Core]AllowUnsafeInputFiles` and `[Core]AllowUnsafeOutputFiles`
    in `miktex.ini` for MiKTeX.  These variables are accessed via
    `latex_config`.

    When reading or writing locations are restricted to the TeX working
    directory plus TEXMF_OUTPUT_DIRECTORY and TEXMFOUTPUT, paths are
    restricted using the following criteria:

      * All relative paths are relative to the TeX working directory.  (This
        is already enforced by `BaseRestrictedPath`.)

      * All absolute paths must be under TEXMF_OUTPUT_DIRECTORY and
        TEXMFOUTPUT.

      * Paths cannot contain `..` to access a parent directory, even if the
        parent directory is a valid location.

    Because paths are analyzed as strings, it is still possible to access
    locations outside the TeX working directory, TEXMF_OUTPUT_DIRECTORY, and
    TEXMFOUTPUT via symlinks in these locations.  Symlinks are not resolved
    in determining whether paths are valid, since paths are analyzed as
    strings without consulting the file system.

    Depending on LaTeX configuration, reading or writing file names starting
    with `.` (dotfiles) may be disabled.

    Under Windows (including Cygwin), writing files with file extensions in
    `PATHEXT` is also disabled.
    '''

    __slots__ = ()

    @classmethod
    def access_file_system_with_resolved_paths(cls):
        return False

    def readable_dir(self) -> tuple[Literal[True], None] | tuple[Literal[False], str]:
        try:
            return self._readable_dir_cache[self.cache_key]
        except KeyError:
            if self.can_read_anywhere():
                self._readable_dir_cache[self.cache_key] = (True, None)
            elif '..' in self.parts:
                self._readable_dir_cache[self.cache_key] = (
                    False,
                    'security settings do not permit paths containing ".."'
                )
            elif self.is_absolute() and not any(self.is_relative_to(p) for p in self.tex_texmfoutput_roots()):
                self._readable_dir_cache[self.cache_key] = (
                    False,
                    'security settings do not permit access to this location'
                )
            else:
                self._readable_dir_cache[self.cache_key] = (True, None)
            return self._readable_dir_cache[self.cache_key]

    def readable_file(self) -> tuple[Literal[True], None] | tuple[Literal[False], str]:
        try:
            return self._readable_file_cache[self.cache_key]
        except KeyError:
            if self.can_read_dotfiles() or not self.name.startswith('.'):
                self._readable_file_cache[self.cache_key] = self.parent.readable_dir()
            else:
                self._readable_file_cache[self.cache_key] = (
                    False,
                    'security settings do not permit access to dotfiles'
                )
            return self._readable_file_cache[self.cache_key]

    def writable_dir(self) -> tuple[Literal[True], None] | tuple[Literal[False], str]:
        try:
            return self._writable_dir_cache[self.cache_key]
        except KeyError:
            if self.can_write_anywhere():
                self._writable_dir_cache[self.cache_key] = (True, None)
            elif '..' in self.parts:
                self._writable_dir_cache[self.cache_key] = (
                    False,
                    'security settings do not permit paths containing ".."'
                )
            elif self.is_absolute() and not any(self.is_relative_to(p) for p in self.tex_texmfoutput_roots()):
                self._writable_dir_cache[self.cache_key] = (
                    False,
                    'security settings do not permit access to this location'
                )
            else:
                self._writable_dir_cache[self.cache_key] = (True, None)
            return self._writable_dir_cache[self.cache_key]

    def writable_file(self) -> tuple[Literal[True], None] | tuple[Literal[False], str]:
        try:
            return self._writable_file_cache[self.cache_key]
        except KeyError:
            prohibited_write_file_extensions = self.prohibited_write_file_extensions()
            if prohibited_write_file_extensions:
                name_lower = self.name.lower()
                for ext in prohibited_write_file_extensions:
                    if name_lower.endswith(ext):
                        self._writable_file_cache[self.cache_key] = (
                            False,
                            f'security settings prevent writing files with extension "{ext}"'
                        )
                        return self._writable_file_cache[self.cache_key]
            if self.can_write_dotfiles() or not self.name.startswith('.'):
                self._writable_file_cache[self.cache_key] = self.parent.writable_dir()
            else:
                self._writable_file_cache[self.cache_key] = (
                    False,
                    'security settings do not permit access to dotfiles'
                )
            return self._writable_file_cache[self.cache_key]


class SafeStringRestrictedPath(StringRestrictedPath):
    '''
    Restrict paths by analyzing them as strings.  TeX configuration is ignored
    and all security settings are at maximum.
    '''

    __slots__ = ()

    @classmethod
    def can_read_anywhere(cls):
        return False

    @classmethod
    def can_read_dotfiles(cls):
        return False

    @classmethod
    def can_write_anywhere(cls):
        return False

    @classmethod
    def can_write_dotfiles(cls):
        return False


class SafeWriteStringRestrictedPath(StringRestrictedPath):
    '''
    Restrict paths by analyzing them as strings.  TeX configuration is ignored
    for writing and all write security settings are at maximum.
    '''

    __slots__ = ()

    @classmethod
    def can_write_anywhere(cls):
        return False

    @classmethod
    def can_write_dotfiles(cls):
        return False




class ResolvedRestrictedPath(BaseRestrictedPath):
    '''
    Restrict paths by resolving any symlinks with the file system and then
    comparing resolved paths to permitted read/write locations.  Restrictions
    are determined by `openout_any` and `openin_any` settings in `texmf.cnf`
    for TeX Live, and from `[Core]AllowUnsafeInputFiles` and
    `[Core]AllowUnsafeOutputFiles` in `miktex.ini` for MiKTeX.  These
    variables are accessed via `latex_config`.

    When reading or writing locations are restricted to the TeX working
    directory plus TEXMF_OUTPUT_DIRECTORY and TEXMFOUTPUT, paths are
    restricted using the following criteria:

      * Resolved paths must be under the TeX working directory,
        TEXMF_OUTPUT_DIRECTORY, or TEXMFOUTPUT.

      * All relative paths are resolved relative to the TeX working directory.
        (This is already enforced by `BaseRestrictedPath`.)

      * Unlike `StringRestrictedPath`, paths are allowed to contain `..`, and
        TEXMF_OUTPUT_DIRECTORY and TEXMFOUTPUT can be accessed via relative
        paths.  This is possible since paths are fully resolved with the file
        system before being compared with permitted read/write locations.

    Because paths are resolved before being compared with permitted read/write
    locations, it is not possible to access locations outside the TeX working
    directory, TEXMF_OUTPUT_DIRECTORY, and TEXMFOUTPUT via symlinks in those
    locations.

    Depending on LaTeX configuration, reading or writing file names starting
    with `.` (dotfiles) may be disabled.

    Under Windows (including Cygwin), writing files with file extensions in
    `PATHEXT` is also disabled.
    '''

    __slots__ = ()

    @classmethod
    def access_file_system_with_resolved_paths(cls):
        return True

    def readable_dir(self) -> tuple[Literal[True], None] | tuple[Literal[False], str]:
        try:
            return self._readable_dir_cache[self.cache_key]
        except KeyError:
            resolved = self.resolve()
            if any(resolved.is_relative_to(p) for p in self.tex_roots_resolved()) or self.can_read_anywhere():
                self._readable_dir_cache[self.cache_key] = (True, None)
            else:
                self._readable_dir_cache[self.cache_key] = (
                    False,
                    'security settings do not permit access to this location'
                )
            return self._readable_dir_cache[self.cache_key]

    def readable_file(self) -> tuple[Literal[True], None] | tuple[Literal[False], str]:
        try:
            return self._readable_file_cache[self.cache_key]
        except KeyError:
            # Must always resolve files in case they are symlinks.  Always use
            # `self.resolve().parent` instead of `self.parent`.
            resolved = self.resolve()
            if not any(p.name.startswith('.') for p in (self, resolved)) or self.can_read_dotfiles():
                self._readable_file_cache[self.cache_key] = resolved.parent.readable_dir()
            else:
                self._readable_file_cache[self.cache_key] = (
                    False,
                    'security settings do not permit access to dotfiles'
                )
            return self._readable_file_cache[self.cache_key]

    def writable_dir(self) -> tuple[Literal[True], None] | tuple[Literal[False], str]:
        try:
            return self._writable_dir_cache[self.cache_key]
        except KeyError:
            resolved = self.resolve()
            if any(resolved.is_relative_to(p) for p in self.tex_roots_resolved()) or self.can_write_anywhere():
                self._writable_dir_cache[self.cache_key] = (True, None)
            else:
                self._writable_dir_cache[self.cache_key] = (
                    False,
                    'security settings do not permit access to this location'
                )
            return self._writable_dir_cache[self.cache_key]

    def writable_file(self) -> tuple[Literal[True], None] | tuple[Literal[False], str]:
        try:
            return self._writable_file_cache[self.cache_key]
        except KeyError:
            resolved = self.resolve()
            prohibited_write_file_extensions = self.prohibited_write_file_extensions()
            if prohibited_write_file_extensions:
                name_lower = self.name.lower()
                resolved_name_lower = resolved.name.lower()
                for ext in prohibited_write_file_extensions:
                    if any(name.endswith(ext) for name in (name_lower, resolved_name_lower)):
                        self._writable_file_cache[self.cache_key] = (
                            False,
                            f'security settings prevent writing files with extension "{ext}"'
                        )
                        return self._writable_file_cache[self.cache_key]
            if not any(p.name.startswith('.') for p in (self, resolved)) or self.can_write_dotfiles():
                self._writable_file_cache[self.cache_key] = resolved.parent.writable_dir()
            else:
                self._writable_file_cache[self.cache_key] = (
                    False,
                    'security settings do not permit access to dotfiles'
                )
            return self._writable_file_cache[self.cache_key]


class SafeResolvedRestrictedPath(ResolvedRestrictedPath):
    '''
    Restrict paths by resolving any symlinks with the file system and then
    comparing resolved paths to permitted read/write locations.  TeX
    configuration is ignored and all security settings are at maximum.
    '''

    __slots__ = ()

    @classmethod
    def can_read_anywhere(cls):
        return False

    @classmethod
    def can_read_dotfiles(cls):
        return False

    @classmethod
    def can_write_anywhere(cls):
        return False

    @classmethod
    def can_write_dotfiles(cls):
        return False


class SafeWriteResolvedRestrictedPath(ResolvedRestrictedPath):
    '''
    Restrict paths by resolving any symlinks with the file system and then
    comparing resolved paths to permitted read/write locations.  TeX
    configuration is ignored for writing and all write security settings are
    at maximum.
    '''

    __slots__ = ()

    @classmethod
    def can_write_anywhere(cls):
        return False

    @classmethod
    def can_write_dotfiles(cls):
        return False
